// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file PetscSolver.h */

#ifndef AMDIS_PETSC_SOLVER_H
#define AMDIS_PETSC_SOLVER_H

#include <set>
#include <map>
#include <mpi.h>

#include "AMDiS_fwd.h"
#include "Global.h"
#include "Parameters.h"
#include "DOFMatrix.h"
#include "parallel/MeshDistributor.h"

#include <petsc.h>
#include <petscsys.h>
#include <petscao.h>

namespace AMDiS {

  using namespace std;

  class PetscSolver
  {
  public:
    PetscSolver()
      : meshDistributor(NULL)
    {}

    virtual ~PetscSolver() {}

    void setMeshDistributor(MeshDistributor *m)
    {
      meshDistributor = m;
    }

    /** \brief
     * Create a PETSc matrix and PETSc vectors. The given DOF matrices are used to
     * create the nnz structure of the PETSc matrix and the values are transfered to it.
     * The given DOF vectors are used to the the values of the PETSc rhs vector.
     *
     * \param[in] mat
     * \param[in] vec
     */
    virtual void fillPetscMatrix(Matrix<DOFMatrix*> *mat, SystemVector *vec) = 0;

    /// Use PETSc to solve the linear system of equations
    virtual void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo) = 0;

  protected:
    MeshDistributor *meshDistributor;
  };


  class PetscSolverGlobalMatrix : public PetscSolver
  {
  public:
    PetscSolverGlobalMatrix()
      : PetscSolver(),
	d_nnz(NULL),
	o_nnz(NULL),
	lastMeshNnz(0),
	zeroStartVector(false)
    {
      GET_PARAMETER(0, "parallel->use zero start vector", "%d", &zeroStartVector);
    }

    void fillPetscMatrix(Matrix<DOFMatrix*> *mat, SystemVector *vec);

    void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo);    

  protected:
    /// Creates a new non zero pattern structure for the PETSc matrix. 
    void createPetscNnzStructure(Matrix<DOFMatrix*> *mat);
    
    /// Takes a DOF matrix and sends the values to the global PETSc matrix.
    void setDofMatrix(DOFMatrix* mat, int dispMult = 1, 
		      int dispAddRow = 0, int dispAddCol = 0);

    /// Takes a DOF vector and sends its values to a given PETSc vector.
    void setDofVector(Vec& petscVec, DOFVector<double>* vec, 
		      int disMult = 1, int dispAdd = 0, bool rankOnly = false);

  protected:
    /// Petsc's matrix structure.
    Mat petscMatrix;

    /** \brief
     * PETSc's vector structures for the rhs vector, the solution vector and a
     * temporary vector for calculating the final residuum.
     */
    Vec petscRhsVec, petscSolVec, petscTmpVec;

    /// Arrays definig the non zero pattern of Petsc's matrix.
    int *d_nnz, *o_nnz;

    /** \brief
     * Stores the mesh change index of the mesh the nnz structure was created for.
     * Therefore, if the mesh change index is higher than this value, we have to create
     * a new nnz structure for PETSc matrices, because the mesh has been changed and
     * therefore also the assembled matrix structure.
     */
    int lastMeshNnz;

    bool zeroStartVector;
  };


#ifdef HAVE_PETSC_DEV
  class PetscSolverSchur : public PetscSolver
  {
  public:
    PetscSolverSchur() 
      : PetscSolver()
    {}

    void fillPetscMatrix(Matrix<DOFMatrix*> *mat, SystemVector *vec);

    void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo);
    
  protected:
    void updateDofData();

    /// Takes a DOF matrix and sends the values to the global PETSc matrix.
    void setDofMatrix(DOFMatrix* mat, int dispMult = 1, 
		      int dispAddRow = 0, int dispAddCol = 0);

    /// Takes a DOF vector and sends its values to a given PETSc vector.
    void setDofVector(Vec& petscVec, DOFVector<double>* vec, 
		      int disMult = 1, int dispAdd = 0, bool rankOnly = false);

  protected:
    int nBoundaryDofs;

    int rStartBoundaryDofs;

    int nOverallBoundaryDofs;

    std::set<DegreeOfFreedom> boundaryDofs;

    map<DegreeOfFreedom, DegreeOfFreedom> mapGlobalBoundaryDof;

    int nInteriorDofs;

    int rStartInteriorDofs;

    int nOverallInteriorDofs;

    std::set<DegreeOfFreedom> interiorDofs;    

    map<DegreeOfFreedom, DegreeOfFreedom> mapGlobalInteriorDof;

    Mat matA11, matA12, matA21, matA22;

    Mat petscMatrix;

    Vec petscRhsVec1, petscRhsVec2;

    Vec petscSolVec1, petscSolVec2;

    Vec petscRhsVec, petscSolVec;
  };
#endif

} // namespace AMDiS

#endif
